@@ -2,20 +2,20 @@ from datetime import datetime
 from unittest.mock import patch
 
 from jdatetime import datetime as jdatetime
-from pandas import DataFrame, DatetimeIndex
+from numpy import dtype
+from pandas import DataFrame, DatetimeIndex, Index
 from pytest import raises
 
+from tsetmc.instruments import Instrument, price_adjustments, search
 # noinspection PyProtectedMember
-from tsetmc import Instrument, _tsetmc, market_watch_init, \
-    closing_price_all, client_type_all, key_stats, \
-    price_adjustments, search
-# noinspection PyProtectedMember
-from tsetmc._tsetmc import _parse_index
+from tsetmc.market_watch import _PRICE_DTYPES, _parse_market_state
+from tsetmc.market_watch import market_watch_init, market_watch_plus, \
+    key_stats, closing_price_all, client_type_all
 
 
-get_patcher = patch.object(
-    _tsetmc, 'get_content', side_effect=NotImplementedError(
-        'tests should not call get_content without patching'))
+get_patcher = patch(
+    'tsetmc._get', side_effect=NotImplementedError(
+        'offline tests should not call tsetmc._get'))
 
 
 def setup_module():
@@ -26,18 +26,23 @@ def teardown_module():
     get_patcher.stop()
 
 
+class FakeResponse:
+
+    def __init__(self, content: bytes):
+        self.content = content
+
+
 def patch_get_content(name):
     with open(f'{__file__}/../testdata/{name}', 'rb') as f:
-        text = f.read()
-    return patch.object(_tsetmc, 'get_content', lambda _: text)
+        content = f.read()
+    return patch('tsetmc._get', lambda _: FakeResponse(content))
 
 
-@patch.object(_tsetmc, 'get_content')
-def test_info_url(mock):
-    with raises(ValueError):
+def test_info_url():
+    with raises(NotImplementedError):
         Instrument(1).info()
-    mock.assert_called_once_with(
-        'http://www.tsetmc.com/tsev2/data/instinfodata.aspx?i=1&c=&e=1')
+    with raises(NotImplementedError):
+        market_watch_init()
 
 
 @patch_get_content('fmelli.html')
@@ -176,13 +181,13 @@ def test_page_info_negative_sector_pe():
 
 @patch_get_content('dara_yekom.txt')
 def test_dara1_instant():
-    assert Instrument(1).info(index=True, orders=True) == {
+    assert Instrument(1).info(market_state=True, orders=True) == {
         'derivatives_status': 'F',
         'derivatives_tno': 7864,
         'derivatives_tval': 150982456491.0,
         'derivatives_tvol': 452251494.0,
         'last_info_datetime': datetime(2021, 1, 27, 12, 30),
-        'market_last_transaction': jdatetime(1399, 11, 8, 15, 21, 59),
+        'market_last_transaction': jdatetime(1499, 11, 8, 15, 21, 59),
         'nav': 190671,
         'nav_datetime': jdatetime(1399, 11, 8, 15, 40),
         'fb_status': 'F',
@@ -301,42 +306,63 @@ def test_fmelli_instant():
         , 'tvol': 66266936}
 
 
+PRICE_DTYPES_ITEMS = [*_PRICE_DTYPES.items()][4:]
+
+
 @patch_get_content('MarketWatchInit.aspx')
 def test_market_watch_init():
-    df = market_watch_init()['dataframe']
-    assert df.dtypes.to_dict() == {
-        'heven': 'int64', 'pf': 'int64', 'pc': 'int64'
-        , 'pl': 'int64', 'tno': 'int64', 'tvol': 'int64'
-        , 'tval': 'int64', 'pmin': 'int64', 'pmax': 'int64'
-        , 'py': 'int64', 'eps': 'float64', 'bvol': 'int64'
-        , 'visitcount': 'int64', 'flow': 'int64'
-        , 'cs': 'int64', 'tmax': 'float64', 'tmin': 'float64'
-        , 'z': 'int64', 'yval': 'int64', 'pd1': 'Int64'
-        , 'po1': 'Int64', 'qd1': 'Int64', 'qo1': 'Int64'
-        , 'zd1': 'Int64', 'zo1': 'Int64', 'pd2': 'Int64'
-        , 'po2': 'Int64', 'qd2': 'Int64', 'qo2': 'Int64'
-        , 'zd2': 'Int64', 'zo2': 'Int64', 'pd3': 'Int64'
-        , 'po3': 'Int64', 'qd3': 'Int64', 'qo3': 'Int64'
-        , 'zd3': 'Int64', 'zo3': 'Int64'}
-    index = df.index
-    assert index.names == ['ins_code', 'isin', 'l18', 'l30']
-    assert index.dtype == 'O'
-    # in pandas 1.3 there will be no need for `.to_frame()`
-    assert index.to_frame().dtypes.to_list() == ['int64', 'O', 'O', 'O']
+    mwi = market_watch_init(join=False, market_state=False)
+    assert [*mwi['prices'].dtypes.items()] == PRICE_DTYPES_ITEMS
+    assert [*mwi['best_limits'].index.dtypes.items()] == [
+        ('ins_code', dtype('uint64')), ('number', dtype('uint64'))]
+    assert 'market_state' not in mwi
+
+    mwi = market_watch_init(market_state=True)
+    prices = mwi['prices']
+    assert 'market_state' in mwi
+    assert 'best_limits' in mwi
+    assert [*prices.dtypes.items()] == [
+        ('pd1', dtype('uint64')),
+        ('po1', dtype('uint64')),
+        ('qd1', dtype('uint64')),
+        ('qo1', dtype('uint64')),
+        ('zd1', dtype('uint64')),
+        ('zo1', dtype('uint64')),
+        ('pd2', dtype('uint64')),
+        ('po2', dtype('uint64')),
+        ('qd2', dtype('uint64')),
+        ('qo2', dtype('uint64')),
+        ('zd2', dtype('uint64')),
+        ('zo2', dtype('uint64')),
+        ('pd3', dtype('uint64')),
+        ('po3', dtype('uint64')),
+        ('qd3', dtype('uint64')),
+        ('qo3', dtype('uint64')),
+        ('zd3', dtype('uint64')),
+        ('zo3', dtype('uint64')),
+        * PRICE_DTYPES_ITEMS]
+
+    assert [*prices.index.dtypes.items()] == [
+        ('ins_code', dtype('uint64')),
+        ('isin', dtype('O')),
+        ('l18', dtype('O')),
+        ('l30', dtype('O'))]
+
+    mwi = market_watch_init(prices=False, market_state=False)
+    assert 'prices' not in mwi
+    assert [*mwi['best_limits'].index.dtypes.items()] == [
+        ('ins_code', dtype('uint64')), ('number', dtype('uint64'))]
 
 
 @patch_get_content('ClosingPriceAll.aspx')
 def test_closing_price_all():
     df = closing_price_all()
-    assert df.dtypes.to_dict() == {
-        'pc': 'int64', 'pl': 'int64', 'tno': 'int64', 'tvol': 'int64'
-        , 'tval': 'int64', 'pmin': 'int64', 'pmax': 'int64', 'py': 'int64'
-        , 'pf': 'int64'}
+    assert all(t == 'uint64' for t in df.dtypes)
+    assert df.columns.to_list() == ['pc', 'pl', 'tno', 'tvol', 'tval', 'pmin', 'pmax', 'py', 'pf']
     index = df.index
     assert index.names == ['ins_code', 'n']
     assert index.dtype == 'O'
-    # in pandas 1.3 there will be no need for `.to_frame()`
-    assert index.to_frame().dtypes.to_list() == ['int64', 'int64']
+    assert all(t == 'uint64' for t in index.dtypes)
 
 
 @patch_get_content('ClientTypeAll.aspx')
@@ -345,7 +371,7 @@ def test_client_type_all():
     assert all(df.columns == [
         'n_buy_count', 'l_buy_count', 'n_buy_volume', 'l_buy_volume'
         , 'n_sell_count', 'l_sell_count', 'n_sell_volume', 'l_sell_volume'])
-    assert all(dt == 'int64' for dt in df.dtypes)
+    assert all(dt == 'uint64' for dt in df.dtypes)
     assert df.index.name == 'ins_code'
 
 
@@ -353,7 +379,9 @@ def test_client_type_all():
 def test_key_stats():
     df = key_stats()
     assert all(df.columns.str.startswith('is'))
+    assert all(t == 'float64' for t in df.dtypes)
     assert df.index.name == 'ins_code'
+    assert df.index.dtype == 'int64'
 
 
 VSKHOOZ = {
@@ -462,12 +490,12 @@ def test_client_type():
 
 def test_parse_index():
     # no tse_value
-    assert _parse_index("00/1/14 06:40:12,F,1294521.64,<div class='mn'>(8671.45)</div>,,0.00,0.00,0,C,0.00,0.00,0,C,0.00,0.00,0,") == {
+    assert _parse_market_state("00/1/14 06:40:12,F,1294521.64,<div class='mn'>(8671.45)</div>,,0.00,0.00,0,C,0.00,0.00,0,C,0.00,0.00,0,") == {
         'derivatives_status': 'C',
         'derivatives_tno': 0,
         'derivatives_tval': 0.0,
         'derivatives_tvol': 0.0,
-        'market_last_transaction': jdatetime(1300, 1, 14, 6, 40, 12),
+        'market_last_transaction': jdatetime(1400, 1, 14, 6, 40, 12),
         'fb_status': 'C',
         'fb_tno': 0,
         'fb_tval': 0.0,
@@ -484,7 +512,7 @@ def test_parse_index():
         'derivatives_tno': 2621,
         'derivatives_tval': 31470939000.0,
         'derivatives_tvol': 101096.0,
-        'market_last_transaction': jdatetime(1399, 12, 16, 15, 45, 46),
+        'market_last_transaction': jdatetime(1400, 12, 16, 15, 45, 46),
         'fb_status': 'N',
         'fb_tno': 342228,
         'fb_tval': 163701347122693.0,
@@ -497,15 +525,15 @@ def test_parse_index():
         'tse_tval': 34288551133025.0,
         'tse_tvol': 3055466451.0,
         'tse_value': 4.674381234630472e+16
-    } == _parse_index(
-        "99/12/16 15:45:46,F,1169760.86,<div class='mn'>(8155.90)</div> -0.69%,46743812346304720.00,3055466451.00,34288551133025.00,428601,N,1057924358.00,163701347122693.00,342228,N,101096.00,31470939000.00,2621,")
+    } == _parse_market_state(
+        "00/12/16 15:45:46,F,1169760.86,<div class='mn'>(8155.90)</div> -0.69%,46743812346304720.00,3055466451.00,34288551133025.00,428601,N,1057924358.00,163701347122693.00,342228,N,101096.00,31470939000.00,2621,")
 
     assert {
         'derivatives_status': 'F',
         'derivatives_tno': 5796,
         'derivatives_tval': 57759844000.0,
         'derivatives_tvol': 225866.0,
-        'market_last_transaction': jdatetime(1399, 12, 24, 14, 39, 40),
+        'market_last_transaction': jdatetime(1400, 12, 24, 14, 39, 40),
         'fb_status': 'F',
         'fb_tno': 544617,
         'fb_tval': 128484547655014.0,
@@ -517,8 +545,8 @@ def test_parse_index():
         'tse_tno': 75828635544957.0,
         'tse_tval': 75828635544957.0,
         'tse_tvol': 9682732949.0,
-        'tse_value': 4.973605456635374e+16} == _parse_index(
-        "99/12/24 14:39:40,F,1245186.04,<div class='pn'>15808.56</div> 1.29%,49736054566353740.00,9682732949.00,75828635544957.00,830860,F,1577202926.00,128484547655014.00,544617,F,225866.00,57759844000.00,5796,")
+        'tse_value': 4.973605456635374e+16} == _parse_market_state(
+        "00/12/24 14:39:40,F,1245186.04,<div class='pn'>15808.56</div> 1.29%,49736054566353740.00,9682732949.00,75828635544957.00,830860,F,1577202926.00,128484547655014.00,544617,F,225866.00,57759844000.00,5796,")
 
 
 @patch_get_content('MarketWatchInit2.aspx')
@@ -645,7 +673,12 @@ def test_adjustments():
 @patch_get_content('adjustments_flow_7.html')
 def test_price_adjustments():
     df = price_adjustments(7)
-    assert df.columns.to_list() == ['l18', 'l30', 'date', 'adj_pc', 'pc']
+    assert [*df.dtypes.items()] == [
+        ('l18', dtype('O')),
+        ('l30', dtype('O')),
+        ('date', dtype('O')),
+        ('adj_pc', dtype('int64')),
+        ('pc', dtype('int64'))]
     assert len(df) == 6
     assert df.iat[-1, -1] == 1000
 
@@ -664,10 +697,94 @@ def test_search():
     df = search('ملت')
     assert type(df) is DataFrame
     assert len(df) == 41
-    assert df.columns.to_list() == [
-        'l18', 'l30', 'ins_code', 'retail', 'compensation', 'wholesale',
-        '_unknown1', '_unknown2', '_unknown3', '_unknown4', '_unknown5']
+    assert [*df.dtypes.items()] == [
+        ('l18', dtype('O')),
+        ('l30', dtype('O')),
+        ('ins_code', dtype('int64')),
+        ('retail', dtype('int64')),
+        ('compensation', dtype('int64')),
+        ('wholesale', dtype('int64')),
+        ('_unknown1', dtype('int64')),
+        ('_unknown2', dtype('int64')),
+        ('_unknown3', dtype('int64')),
+        ('_unknown4', dtype('int64')),
+        ('_unknown5', dtype('O'))]
 
 
 def test_l18_without_web_request():
     assert Instrument(46348559193224090).l18 == 'فولاد'
+
+
+@patch_get_content('MarketWatchPlus00.txt')
+def test_market_watch_plus_new():
+    mwp = market_watch_plus(0, 0, messages=False, market_state=False)
+    new_prices = mwp['new_prices']
+    assert [*new_prices.dtypes.items()] == PRICE_DTYPES_ITEMS
+    assert [*new_prices.index.dtypes.items()] == [
+        ('ins_code', dtype('uint64')),
+        ('isin', dtype('O')),
+        ('l18', dtype('O')),
+        ('l30', dtype('O'))]
+    best_limits = mwp['best_limits']
+    assert all(t == 'uint64' for t in best_limits.dtypes)
+    assert best_limits.columns.to_list() == ['number', 'zo', 'zd', 'pd', 'po', 'qd', 'qo']
+    assert best_limits.index.dtype == dtype('uint64')  # ins_code
+    assert 'messages' not in mwp
+    assert 'market_state' not in mwp
+
+
+@patch_get_content('MarketWatchPlus_h64130_r9540883525.txt')
+def test_market_watch_plus_update():
+    mwp = market_watch_plus(64130, 9540883525)
+
+    price_updates = mwp['price_updates']
+    assert [*price_updates.dtypes.items()] == [
+        ('heven', dtype('uint64')),
+        ('pf', dtype('uint64')),
+        ('pc', dtype('uint64')),
+        ('pl', dtype('uint64')),
+        ('tno', dtype('uint64')),
+        ('tvol', dtype('uint64')),
+        ('tval', dtype('uint64')),
+        ('pmin', dtype('uint64')),
+        ('pmax', dtype('uint64'))]
+    assert price_updates.index.dtype == 'uint64'
+
+    assert mwp['market_state'] == {
+        'market_last_transaction': jdatetime(1400, 9, 21, 6, 41, 30),
+        'tse_status': 'F',
+        'tse_index': 1344441.58,
+        'tse_index_change': -4948.21,
+        'tse_tvol': 0.0,
+        'tse_tval': 0.0,
+        'tse_tno': 0.0,
+        'fb_status': 'F',
+        'fb_tvol': 0.0,
+        'fb_tval': 0.0,
+        'fb_tno': 0,
+        'derivatives_status': 'F',
+        'derivatives_tvol': 0.0,
+        'derivatives_tval': 0.0,
+        'derivatives_tno': 0,
+        'tse_index_change_percent': -0.37}
+
+    assert mwp['messages'] == ['115048', '770345', '427827']
+
+    best_limits = mwp['best_limits']
+    assert [*best_limits.dtypes.items()] == [
+        ('number', dtype('uint64')),
+        ('zo', dtype('uint64')),
+        ('zd', dtype('uint64')),
+        ('pd', dtype('uint64')),
+        ('po', dtype('uint64')),
+        ('qd', dtype('uint64')),
+        ('qo', dtype('uint64'))]
+    assert best_limits.index.dtype == 'uint64'
+
+    assert mwp['refid'] == 9540883545
+
+    new_prices = mwp['new_prices']
+    assert [*new_prices.dtypes.items()] == PRICE_DTYPES_ITEMS
+    assert [*new_prices.index.dtypes.items()] == [
+        ('ins_code', dtype('uint64')),
+        ('isin', dtype('O')), ('l18', dtype('O')), ('l30', dtype('O'))]
